import { _decorator, Label, UITransform, Node, v3, Tween, tween, UIOpacity } from "cc";
import { PanelConfigs } from "../../../mainbundle/scripts/configs/PanelConfigs";
import { gg } from "../../../scripts/framework/gg";
import { PanelComponent, PanelShowOption, PanelHideOption } from "../../../scripts/framework/lib/router/PanelComponent";

const { ccclass, property } = _decorator;

export interface ToastPanelShowArgs {
    /**
     * 文本内容
     */
    text: string;

    /**
     * Toast展示持续时间(s)，不传则默认显示2秒后自动消失
     */
    duration?: number;
}

/**
 * Toast 面板
 */
@ccclass('ToastPanelPrefab')
export default class ToastPanelPrefab extends PanelComponent {
    @property(Node)
    textBgNode: Node = null;

    @property(Label)
    textLabel: Label = null;

    @property({ tooltip: "文本上边界距离文本背景节点上边界的Margin值" })
    textMarginTop: number = 30;

    @property({ tooltip: "文本下边界距离文本背景节点下边界的Margin值" })
    textMarginBottom: number = 30;

    @property({ tooltip: "文本左边界距离文本背景节点左边界的Margin值" })
    textMarginLeft: number = 30;

    @property({ tooltip: "文本右边界距离文本背景节点右边界的Margin值" })
    textMarginRight: number = 30;

    show(option: PanelShowOption): void {
        let showArgs = option.data as ToastPanelShowArgs;
        let duration = showArgs.duration == null ? 2 : showArgs.duration;
        this._updateToastLabel(showArgs.text);
        this._startHideScheduler(duration);
        this._playPanelShowAnim();
        option.onShowed();
    }

    private _updateToastLabel(text: string) {
        // 设置文本最大长度
        let textLabelMaxWidth = this.node.getComponent(UITransform).width * 0.6;

        // NONE 模式渲染文本，得到文本宽度
        this.textLabel.overflow = Label.Overflow.NONE;
        this.textLabel.string = text;
        if (this.textLabel["_forceUpdateRenderData"] != null) {
            this.textLabel["_forceUpdateRenderData"]();
        }

        if (this.textLabel.node.getComponent(UITransform).width > textLabelMaxWidth) {
            // 如果大于最大宽度，那么设置为 RESIZE_HEIGHT 模式重新渲染
            this.textLabel.node.getComponent(UITransform).width = textLabelMaxWidth;
            this.textLabel.overflow = Label.Overflow.RESIZE_HEIGHT;
            this.textLabel.string = text;
            if (this.textLabel["_forceUpdateRenderData"] != null) {
                this.textLabel["_forceUpdateRenderData"]();
            }
        }

        // 根据文本宽高，重新设置文本背景节点的宽高
        setTimeout(() => {
            this.textBgNode.getComponent(UITransform).width = this.textLabel.node.getComponent(UITransform).width + this.textMarginLeft + this.textMarginRight;
            this.textBgNode.getComponent(UITransform).height = this.textLabel.node.getComponent(UITransform).height + this.textMarginTop + this.textMarginBottom;
        }, 10);
    }

    /**
     * 启动隐藏倒计时
     */
    private _startHideScheduler(duration: number) {
        this.unscheduleAllCallbacks();
        this.scheduleOnce(() => {
            gg.panelRouter.hide({
                panel: PanelConfigs.toastPanel,
            });
        }, duration);
    }

    hide(option: PanelHideOption): void {
        this.unscheduleAllCallbacks();
        this._playPanelHideAnim(() => {
            option.onHided();
        });
    }

    // ////////////////////////////////////////////////////////////
    // 面板动画相关

    private _playPanelShowAnim(onAnimCompleted?: Function) {
        Tween.stopAllByTarget(this.textBgNode);
        tween<Node>(this.textBgNode)
            .set({
                scale: v3(),
                position: v3(0, -100, 0),
            })
            .to(0.24, { scale: v3(1, 1), position: v3() }, { easing: "backOut" })
            .call(() => {
                onAnimCompleted && onAnimCompleted();
            })
            .start();

        Tween.stopAllByTarget(this.textBgNode.getComponent(UIOpacity));
        tween<UIOpacity>(this.textBgNode.getComponent(UIOpacity))
            .set({
                opacity: 0,
            })
            .to(0.24, { opacity: 255 }, { easing: "backOut" })
            .call(() => {
                onAnimCompleted && onAnimCompleted();
            })
            .start();
    }

    private _playPanelHideAnim(onHideComleted: Function) {
        Tween.stopAllByTarget(this.textBgNode.getComponent(UIOpacity));
        tween(this.textBgNode.getComponent(UIOpacity))
            .to(0.24, { opacity: 0 })
            .call(() => {
                onHideComleted();
            })
            .start();
    }
}
